import App
import FoundationTech
import loadspacehelper
import math
import MissionLib


"""

Sample Setup:

Foundation.ShipDef.VasKholhr.dTechs = { 'SubModel': {
        "Setup":        {
                "Body":                 "VasKholhr_Body",
                "NormalModel":          shipFile,
                "WarpModel":          "VasKholhr_WingUp",
                "AttackModel":          "VasKholhr_WingDown",
                "Hardpoints":       {
                        "Port Cannon":  [-0.677745, 0.514529, -0.229285],
                        "Star Cannon":  [0.663027, 0.511252, -0.240265],
                        "Port Cannon 1":  [-0.323324, 0.240263, -0.115398],
                        "Star Cannon 1":  [0.319566, 0.242142, -0.11861],
                },
                "AttackHardpoints":       {
                        "Port Cannon":  [-0.503543, 0.524792, -0.47761],
                        "Star Cannon":  [0.486256, 0.527008, -0.483889],
                        "Port Cannon 1":  [-0.244469, 0.228191, -0.19762],
                        "Star Cannon 1":  [0.243789, 0.243208, -0.201933],
                },
        },
                
        "Port Wing":     ["VasKholhr_Portwing", {
                "Position":             [0, 0, 0],
                "Rotation":             [0, 0, 0], # normal Rotation used if not Red Alert and if not Warp
                "AttackRotation":         [0, -0.6, 0],
                "AttackDuration":         200.0, # Value is 1/100 of a second
                "AttackPosition":         [0, 0, 0.03],
                "WarpRotation":       [0, 0.349, 0],
                "WarpPosition":       [0, 0, 0.02],
                "WarpDuration":       150.0,
                }
        ],
        
        "Starboard Wing":     ["VasKholhr_Starboardwing", {
                "Position":             [0, 0, 0],
                "Rotation":             [0, 0, 0],
                "AttackRotation":         [0, 0.6, 0],
                "AttackDuration":         200.0, # Value is 1/100 of a second
                "AttackPosition":         [0, 0, 0.03],
                "WarpRotation":       [0, -0.349, 0],
                "WarpPosition":       [0, 0, 0.02],
                "WarpDuration":       150.0,
                }
        ],
}}



"""

# This class does control the attach and detach of the Models
class SubModels(FoundationTech.TechDef):
	
	# called by FoundationTech when a ship is created
	# Prepares the ship for moving its sub parts
        def AttachShip(self, pShip, pInstance):
                print "Ship %s with SubModels support added" % pShip.GetName()

                sNamePrefix = pShip.GetName() + "_"
                pInstance.__dict__["OptionsList"] = [] # save options to this list, so we can access them later
                self.bAddedWarpListener = {} # this variable will make sure we add our event handlers only once
                self.bAddedAlertListener = {}
                ModelList = pInstance.__dict__["SubModel"]
                
                if not ModelList.has_key("Setup"):
                        print "Error: Cannot find Setup for Moving Parts"
                        return
                
		# Interate over every item
                for sNameSuffix in ModelList.keys():
			# if this is the setup, do the thinks for the whole ship 
			#we have to remember, like alert state
                        if sNameSuffix == "Setup":
                                dOptions = ModelList[sNameSuffix]
                                # we start with green alert
                                if not dOptions.has_key("AlertLevel"):
                                        dOptions["AlertLevel"] = {}
                                if not dOptions.has_key("GenMoveID"):
                                        dOptions["GenMoveID"] = {}
                                dOptions["AlertLevel"][pShip.GetName()] = 0
                                dOptions["GenMoveID"][pShip.GetName()] = 0
                                pInstance.OptionsList.append("Setup", dOptions)
                                continue # nothing more todo here
                        #
                        # the following stuff is only for the objects that moves
                        
                        if len(ModelList[sNameSuffix]) > 1:
                                dOptions = ModelList[sNameSuffix][1]
                        sFile = ModelList[sNameSuffix][0]
                        loadspacehelper.PreloadShip(sFile, 1)
                        
			# save the shipfile for later use
                        dOptions["sShipFile"] = sFile
                        
                        # set current rotation/position values
                        if not dOptions.has_key("currentRotation"):
                                dOptions["currentRotation"] = {}
                        if not dOptions.has_key("currentPosition"):
                                dOptions["currentPosition"] = {}
                        if not dOptions.has_key("curMovID"):
                                dOptions["curMovID"] = {}
                        dOptions["currentRotation"][pShip.GetName()] = [0, 0, 0] # those two lists are updated with every move
                        dOptions["currentPosition"][pShip.GetName()] = [0, 0, 0]
                        dOptions["curMovID"][pShip.GetName()] = 0
                        if not dOptions.has_key("Position"):
                                dOptions["Position"] = [0, 0, 0]
                        dOptions["currentPosition"][pShip.GetName()] = dOptions["Position"] # set current Position
                        
                        if not dOptions.has_key("Rotation"):
                                dOptions["Rotation"] = [0, 0, 0]
                        dOptions["currentRotation"][pShip.GetName()] = dOptions["Rotation"] # set current Rotation
                        
                        # event listener
                        if not self.bAddedWarpListener.has_key(pShip.GetName()) and (dOptions.has_key("WarpRotation") or dOptions.has_key("WarpPosition")):
                                pShip.AddPythonFuncHandlerForInstance(App.ET_START_WARP_NOTIFY, __name__ + ".StartingWarp")
                                # ET_EXITED_WARP handler doesn't seem to work, so use ET_EXITED_SET instead
                                pShip.AddPythonFuncHandlerForInstance(App.ET_EXITED_SET, __name__ + ".ExitSet")
                                self.bAddedWarpListener[pShip.GetName()] = 1
                        if not self.bAddedAlertListener.has_key(pShip.GetName()) and (dOptions.has_key("AttackRotation") or dOptions.has_key("AttackPosition")):
                                # Alert change handler doesn't work for AI ships, so use subsystem changed instead
                                pShip.AddPythonFuncHandlerForInstance(App.ET_SUBSYSTEM_STATE_CHANGED, __name__ + ".SubsystemStateChanged")
                                self.bAddedAlertListener[pShip.GetName()] = 1
                
                # Make sure the Ship is correctly set
                # because we don't get the first ET_SUBSYSTEM_STATE_CHANGED event for Ai ships
                PartsForWeaponState(pShip)

	# Called by FoundationTech when a Ship is removed from set (eg destruction)
        def DetachShip(self, iShipID, pInstance):
		# get our Ship
                pShip = App.ShipClass_GetObjectByID(None, iShipID)
                if pShip:
			# remove the listeners
                        if self.bAddedWarpListener.has_key(pShip.GetName()):
                                pShip.RemoveHandlerForInstance(App.ET_START_WARP_NOTIFY, __name__ + ".StartingWarp")
                                pShip.RemoveHandlerForInstance(App.ET_EXITED_SET, __name__ + ".ExitSet")
                                del self.bAddedWarpListener[pShip.GetName()]
                        if self.bAddedAlertListener.has_key(pShip.GetName()):
                                pShip.RemoveHandlerForInstance(App.ET_SUBSYSTEM_STATE_CHANGED, __name__ + ".SubsystemStateChanged")
                                del self.bAddedAlertListener[pShip.GetName()]
			
                        if hasattr(pInstance, "OptionsList"):
                                for item in pInstance.OptionsList:
                                        if item[0] == "Setup":
                                                del item[1]["AlertLevel"][pShip.GetName()]
                                        else:
                                                del item[1]["currentRotation"][pShip.GetName()]
                                                del item[1]["currentPosition"][pShip.GetName()]
                                
                if hasattr(pInstance, "SubModelList"):
                        del pInstance.SubModelList

	# Attaches the SubParts to the Body Model
        def AttachParts(self, pShip, pInstance):
                pSet = pShip.GetContainingSet()
                pInstance.__dict__["SubModelList"] = []
                ModelList = pInstance.__dict__["SubModel"]
                sNamePrefix = pShip.GetName() + "_"
                SubModelList = pInstance.SubModelList

		# iteeeerate over every SubModel
                for sNameSuffix in ModelList.keys():
                        if sNameSuffix == "Setup":
                                continue

                        sFile = ModelList[sNameSuffix][0]
                        sShipName = sNamePrefix + sNameSuffix
                        
                        # check if the ship does exist first, before create it
                        pSubShip = MissionLib.GetShip(sShipName)
                        if not pSubShip:
                                pSubShip = loadspacehelper.CreateShip(sFile, pSet, sShipName, "")
                        SubModelList.append(pSubShip)
                        if len(ModelList[sNameSuffix]) > 1:
                                dOptions = ModelList[sNameSuffix][1]

                        # set current positions
                        pSubShip.SetTranslateXYZ(dOptions["currentPosition"][pShip.GetName()][0],dOptions["currentPosition"][pShip.GetName()][1],dOptions["currentPosition"][pShip.GetName()][2])
                        iNorm = math.sqrt(dOptions["currentRotation"][pShip.GetName()][0] ** 2 + dOptions["currentRotation"][pShip.GetName()][1] ** 2 + dOptions["currentRotation"][pShip.GetName()][2] ** 2)
                        pSubShip.SetAngleAxisRotation(iNorm,dOptions["currentRotation"][pShip.GetName()][0],dOptions["currentRotation"][pShip.GetName()][1],dOptions["currentRotation"][pShip.GetName()][2])

                        # save the options list
                        iSaveDone = 0
                        # this is here to check if we already have the entry
                        for lList in pInstance.OptionsList:
                                if lList[0] != "Setup":
                                        if lList[1]["sShipFile"] == sFile:
                                                lList[0] = pSubShip
                                                iSaveDone = 1
                                                break
                        
                        if not iSaveDone:
                                pInstance.OptionsList.append([pSubShip, dOptions])
			
                        pSubShip.SetUsePhysics(0)
                        pSubShip.SetTargetable(0)
                        #pSubShip.SetInvincible(1)
                        #pSubShip.SetHurtable(0)
                        #pSubShip.GetShipProperty().SetMass(1.0e+25)
                        #pSubShip.GetShipProperty().SetRotationalInertia(1.0e+25)
                        #pSubShip.GetShipProperty().SetStationary(1)
                        pSubShip.SetHailable(0)
                        if pSubShip.GetShields():
                                pSubShip.GetShields().TurnOff()
            
                        pShip.EnableCollisionsWith(pSubShip, 0)
                        pSubShip.EnableCollisionsWith(pShip, 0)
                        for pSubShip2 in SubModelList:
                                pSubShip.EnableCollisionsWith(pSubShip2, 0)
                                pSubShip2.EnableCollisionsWith(pSubShip, 0)                        
                        
                        pShip.AttachObject(pSubShip)

	# check if parts are attached
        def ArePartsAttached(self, pShip, pInstance):
                if hasattr(pInstance, "SubModelList"):
                        return 1
                return 0

	# Detaches the parts
        def DetachParts(self, pShip, pInstance):
                if hasattr(pInstance, "SubModelList"):
                        for pSubShip in pInstance.SubModelList:
                                pSet = pSubShip.GetContainingSet()
                                pShip.DetachObject(pSubShip)
                                pSet.DeleteObjectFromSet(pSubShip.GetName())
                        del pInstance.SubModelList
                
oSubModels = SubModels("SubModel")


# The class does the moving of the parts
# with every move the part continues to move
class MovingEvent:
	# prepare fore move...
        def __init__(self, pShip, item, fDuration, lStartingRotation, lStoppingRotation, lStartingTranslation, lStoppingTranslation):
                self.iNacelleID = item[0].GetObjID()
                self.iThisMovID = item[1]["curMovID"][pShip.GetName()]
                self.dOptionsList = item[1]
                self.pShip = pShip
                
                # rotation values
                self.iCurRotX = lStartingRotation[0]
                self.iCurRotY = lStartingRotation[1]
                self.iCurRotZ = lStartingRotation[2]
		if fDuration > 0:
			self.iRotStepX = (lStoppingRotation[0] - lStartingRotation[0]) / (fDuration * 1.0) # the 1.0 makes sure we have a float
			self.iRotStepY = (lStoppingRotation[1] - lStartingRotation[1]) / (fDuration * 1.0)
			self.iRotStepZ = (lStoppingRotation[2] - lStartingRotation[2]) / (fDuration * 1.0)
		else:
			self.iRotStepX = (lStoppingRotation[0] - lStartingRotation[0])
			self.iRotStepY = (lStoppingRotation[1] - lStartingRotation[1])
			self.iRotStepZ = (lStoppingRotation[2] - lStartingRotation[2])
                
                # translation values
                self.iCurTransX = lStartingTranslation[0]
                self.iCurTransY = lStartingTranslation[1]
                self.iCurTransZ = lStartingTranslation[2]
		if fDuration > 0:
			self.iTransStepX = (lStoppingTranslation[0] - lStartingTranslation[0]) / (fDuration * 1.0)
			self.iTransStepY = (lStoppingTranslation[1] - lStartingTranslation[1]) / (fDuration * 1.0)
			self.iTransStepZ = (lStoppingTranslation[2] - lStartingTranslation[2]) / (fDuration * 1.0)
		else:
			self.iTransStepX = (lStoppingTranslation[0] - lStartingTranslation[0])
			self.iTransStepY = (lStoppingTranslation[1] - lStartingTranslation[1])
			self.iTransStepZ = (lStoppingTranslation[2] - lStartingTranslation[2])


	# move!
        def __call__(self):
		# if the move ID doesn't match then this move is outdated
                if self.iThisMovID != self.dOptionsList["curMovID"][self.pShip.GetName()]:
                        print "Moving Error: Move no longer active"
                        return 1
                
                # this makes sure the game does not crash when trying to access a deleted element
                pNacelle = App.ShipClass_GetObjectByID(None, self.iNacelleID)
                if not pNacelle:
                        print "Moving Error: Lost part"
                        return 0
                        
                # set new Rotation values
                self.iCurRotX = self.iCurRotX + self.iRotStepX
                self.iCurRotY = self.iCurRotY + self.iRotStepY
                self.iCurRotZ = self.iCurRotZ + self.iRotStepZ
                iNorm = math.sqrt(self.iCurRotX ** 2 + self.iCurRotY ** 2 + self.iCurRotZ ** 2)
                # set Rotation
                pNacelle.SetAngleAxisRotation(iNorm, self.iCurRotX, self.iCurRotY, self.iCurRotZ)

                # set new Translation values
                self.iCurTransX = self.iCurTransX + self.iTransStepX
                self.iCurTransY = self.iCurTransY + self.iTransStepY
                self.iCurTransZ = self.iCurTransZ + self.iTransStepZ
                # set Translation
                pNacelle.SetTranslateXYZ(self.iCurTransX, self.iCurTransY, self.iCurTransZ)
                
                self.dOptionsList["currentRotation"][self.pShip.GetName()] = [self.iCurRotX, self.iCurRotY, self.iCurRotZ]
                self.dOptionsList["currentPosition"][self.pShip.GetName()] = [self.iCurTransX, self.iCurTransY, self.iCurTransZ]
                
                return 0


# calls the MovingEvent class and returns its return value
def MovingAction(pAction, oMovingEvent):
        return oMovingEvent()


# called when a ship changes Power of one of its subsystems
# cause this is possible also an alert event
def SubsystemStateChanged(pObject, pEvent):
        pShip = App.ShipClass_Cast(pObject)
        pSubsystem = pEvent.GetSource()

	# if the subsystem that changes its power is a weapon
        if pSubsystem.IsTypeOf(App.CT_WEAPON_SYSTEM):
		# set wings for this alert state
                PartsForWeaponState(pShip)
		
        pObject.CallNextHandler(pEvent)
        

# called when a ship exits a Set. Replacement for WARP_END Handler.
def ExitSet(pObject, pEvent):
        pShip   = App.ShipClass_Cast(pEvent.GetDestination())
        sSetName = pEvent.GetCString()
	# if the system we come from is the warp system, then we exitwarp, right?
        if sSetName == "warp":
		# call ExitingWarp in a few seconds
                pSeq = App.TGSequence_Create()
                pSeq.AppendAction(App.TGScriptAction_Create(__name__, "ExitingWarp", pShip), 4.0)
                pSeq.Play()
		
        pObject.CallNextHandler(pEvent)


# Replaces the Model of pShip
def ReplaceModel(pShip, sNewShipScript):
        ShipScript = __import__(('ships.' + sNewShipScript))
        ShipScript.LoadModel()
        kStats = ShipScript.GetShipStats()
        pShip.SetupModel(kStats['Name'])


# Prepares a ship to move: Replaces the current Model with the move Model and attaches its sub Models
def PrepareShipForMove(pShip, pInstance):
        if not oSubModels.ArePartsAttached(pShip, pInstance):
                ReplaceModel(pShip, pInstance.__dict__["SubModel"]["Setup"]["Body"])
                oSubModels.AttachParts(pShip, pInstance)


def IncCurrentMoveID(pShip, pInstance):
        for item in pInstance.OptionsList:
                if item[0] == "Setup":
                        item[1]["GenMoveID"][pShip.GetName()] = item[1]["GenMoveID"][pShip.GetName()] + 1


def GetCurrentMoveID(pShip, pInstance):
        iGenMoveID = 0
        
        for item in pInstance.OptionsList:
                if item[0] == "Setup":
                        iGenMoveID = item[1]["GenMoveID"][pShip.GetName()]
        return iGenMoveID
                        

def MoveFinishMatchId(pShip, pInstance, iThisMovID):
        if GetCurrentMoveID(pShip, pInstance) == iThisMovID:
                return 1
        return 0
        
        
# called after the Alert move action
# Remove the attached parts and use the attack or normal model now
def AlertMoveFinishAction(pAction, pShip, pInstance, iThisMovID):
        
        # Don't switch Models back when the ID does not match
        if not MoveFinishMatchId(pShip, pInstance, iThisMovID):
                return 1
        
        oSubModels.DetachParts(pShip, pInstance)
        if pShip.GetAlertLevel() == 2:
                sNewShipScript = pInstance.__dict__["SubModel"]["Setup"]["AttackModel"]
        else:
                sNewShipScript = pInstance.__dict__["SubModel"]["Setup"]["NormalModel"]
        ReplaceModel(pShip, sNewShipScript)
        
        return 0


# called after the warp-start move action
# Remove the attached parts and use the warp model now
def WarpStartMoveFinishAction(pAction, pShip, pInstance, iThisMovID):
        # Don't switch Models back when the ID does not match
        if not MoveFinishMatchId(pShip, pInstance, iThisMovID):
                return 1
                
        oSubModels.DetachParts(pShip, pInstance)
        sNewShipScript = pInstance.__dict__["SubModel"]["Setup"]["WarpModel"]
        ReplaceModel(pShip, sNewShipScript)
        return 0


# called after the warp-exit move action
# Remove the attached parts and use the attack or normal model now
def WarpExitMoveFinishAction(pAction, pShip, pInstance, iThisMovID):
        # Don't switch Models back when the ID does not match
        if not MoveFinishMatchId(pShip, pInstance, iThisMovID):
                return 1
                
        oSubModels.DetachParts(pShip, pInstance)
        if pInstance.__dict__["SubModel"]["Setup"].has_key("AttackModel") and pShip.GetAlertLevel() == 2:
                sNewShipScript = pInstance.__dict__["SubModel"]["Setup"]["AttackModel"]
        else:
                sNewShipScript = pInstance.__dict__["SubModel"]["Setup"]["NormalModel"]
        ReplaceModel(pShip, sNewShipScript)
        return 0
        

# called after a move, sets the current Rotation/Translation values to the final ones
def UpdateState(pAction, pShip, item, lStoppingRotation, lStoppingTranslation):
        item[1]["currentRotation"][pShip.GetName()] = lStoppingRotation
        item[1]["currentPosition"][pShip.GetName()] = lStoppingTranslation
        item[1]["curMovID"][pShip.GetName()] = 0
        return 0
        
        
def UpdateHardpointPositions(pAction, pShip, dHardpoints):
        for sHP in dHardpoints.keys():
                pHP = MissionLib.GetSubsystemByName(pShip, sHP)
                pHPprob = pHP.GetProperty()
                pHPprob.SetPosition(dHardpoints[sHP][0], dHardpoints[sHP][1], dHardpoints[sHP][2])
        pShip.SetupProperties()
        
        return 0


# Set the parts to the correct alert state
def PartsForWeaponState(pShip):
        pInstance = FoundationTech.dShips[pShip.GetName()]
        iType = pShip.GetAlertLevel()
        iLongestTime = 0.0
        dHardpoints = {}
        
        # check if ship still exits
        pShip = App.ShipClass_GetObjectByID(None, pShip.GetObjID())
        if not pShip:
                return
        
        # try to get the last alert level
        for item in pInstance.OptionsList:
                if item[0] == "Setup":
                        dGenShipDict = item[1]
                        break
	
	# check if the alert state has really chanced since the last time
        if dGenShipDict["AlertLevel"][pShip.GetName()] == iType:
                return
	# update alert state
        dGenShipDict["AlertLevel"][pShip.GetName()] = iType
        IncCurrentMoveID(pShip, pInstance)
        
        # start with replacing the Models
        PrepareShipForMove(pShip, pInstance)
        
	# iterate over every submodel
        for item in pInstance.OptionsList:
                if item[0] == "Setup":
                        # attack or cruise modus?
                        if iType == 2 and item[1].has_key("AttackHardpoints"):
                                dHardpoints = item[1]["AttackHardpoints"]
                        elif iType != 2 and item[1].has_key("Hardpoints"):
                                dHardpoints = item[1]["Hardpoints"]
                        
                        # setup is not a submodel
                        continue
        
		# set the id for this move
                iThisMovID = item[1]["curMovID"][pShip.GetName()] + 1
                item[1]["curMovID"][pShip.GetName()] = iThisMovID
        
                fDuration = 200.0
                if item[1].has_key("AttackDuration"):
                        fDuration = item[1]["AttackDuration"]
                    
                # Rotation
                lStartingRotation = item[1]["currentRotation"][pShip.GetName()]
                lStoppingRotation = lStartingRotation
                if item[1].has_key("AttackRotation") and iType == 2:
                        lStoppingRotation = item[1]["AttackRotation"]
		else:
			lStoppingRotation = item[1]["Rotation"]
                
		
                # Translation
                lStartingTranslation = item[1]["currentPosition"][pShip.GetName()]
                lStoppingTranslation = lStartingTranslation
                if item[1].has_key("AttackPosition") and iType == 2:
                        lStoppingTranslation = item[1]["AttackPosition"]
		else:
			lStoppingTranslation = item[1]["Position"]
        
                iTime = 0.0
                iTimeNeededTotal = 0.0
                oMovingEvent = MovingEvent(pShip, item, fDuration, lStartingRotation, lStoppingRotation, lStartingTranslation, lStoppingTranslation)
                pSeq = App.TGSequence_Create()
                
		# do the move
                while(iTime < fDuration):
                        if iTime == 0.0:
                                iWait = 0.5 # we wait for the first run
                        else:
                                iWait = 0.01 # normal step
                        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "MovingAction", oMovingEvent), iWait)
                        iTimeNeededTotal = iTimeNeededTotal + iWait
                        iTime = iTime + 1
                pSeq.AppendAction(App.TGScriptAction_Create(__name__, "UpdateState", pShip, item, lStoppingRotation, lStoppingTranslation))
                pSeq.Play()
                
                # iLongestTime is for the part that needs the longest time...
                if iTimeNeededTotal > iLongestTime:
                        iLongestTime = iTimeNeededTotal
                
        # finally detach
        pSeq = App.TGSequence_Create()
        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "UpdateHardpointPositions", pShip, dHardpoints), iLongestTime)
        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "AlertMoveFinishAction", pShip, pInstance, GetCurrentMoveID(pShip, pInstance)), 1.0)
        pSeq.Play()


# Set the parts for Warp state
def StartingWarp(pObject, pEvent):
        pShip = App.ShipClass_Cast(pObject)
        pInstance = FoundationTech.dShips[pShip.GetName()]
        iLongestTime = 0.0
        IncCurrentMoveID(pShip, pInstance)
        dHardpoints = {}
        
        # first replace the Models
        PrepareShipForMove(pShip, pInstance)
        
        for item in pInstance.OptionsList:
		# setup is not a submodel
                if item[0] == "Setup":
                        if item[1].has_key("WarpHardpoints"):
                                dHardpoints = item[1]["WarpHardpoints"]
                        continue
        
		# set the id for this move
                iThisMovID = item[1]["curMovID"][pShip.GetName()] + 1
                item[1]["curMovID"][pShip.GetName()] = iThisMovID
        
                fDuration = 200.0
                if item[1].has_key("WarpDuration"):
                        fDuration = item[1]["WarpDuration"]
                    
                # Rotation
                lStartingRotation = item[1]["currentRotation"][pShip.GetName()]
                lStoppingRotation = lStartingRotation
                if item[1].has_key("WarpRotation"):
                        lStoppingRotation = item[1]["WarpRotation"]
                
                # Translation
                lStartingTranslation = item[1]["currentPosition"][pShip.GetName()]
                lStoppingTranslation = lStartingTranslation
                if item[1].has_key("WarpPosition"):
                        lStoppingTranslation = item[1]["WarpPosition"]
        
                iTime = 0.0
                iTimeNeededTotal = 0.0
                oMovingEvent = MovingEvent(pShip, item, fDuration, lStartingRotation, lStoppingRotation, lStartingTranslation, lStoppingTranslation)
                pSeq = App.TGSequence_Create()
                
		# do the move
                while(iTime < fDuration):
                        if iTime == 0.0:
                                iWait = 0.5 # we wait for the first run
                        else:
                                iWait = 0.01 # normal step
                        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "MovingAction", oMovingEvent), iWait)
                        iTimeNeededTotal = iTimeNeededTotal + iWait
                        iTime = iTime + 1
                pSeq.AppendAction(App.TGScriptAction_Create(__name__, "UpdateState", pShip, item, lStoppingRotation, lStoppingTranslation))
                pSeq.Play()
                
                # iLongestTime is for the part that needs the longest time...
                if iTimeNeededTotal > iLongestTime:
                        iLongestTime = iTimeNeededTotal
                
        # finally detach
        pSeq = App.TGSequence_Create()
        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "UpdateHardpointPositions", pShip, dHardpoints), iLongestTime)
        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "WarpStartMoveFinishAction", pShip, pInstance, GetCurrentMoveID(pShip, pInstance)), 1.0)
        pSeq.Play()

        pObject.CallNextHandler(pEvent)


def ExitingWarp(pAction, pShip):
        pInstance = FoundationTech.dShips[pShip.GetName()]
        iLongestTime = 0.0
        IncCurrentMoveID(pShip, pInstance)
        dHardpoints = {}
        
        # first replace the Models
        PrepareShipForMove(pShip, pInstance)
        
        for item in pInstance.OptionsList:
		# setup is not a submodel
                if item[0] == "Setup":
                        if item[1].has_key("WarpHardpoints"):
                                dHardpoints = item[1]["WarpHardpoints"]
                        continue
        
		# set the id for this move
                iThisMovID = item[1]["curMovID"][pShip.GetName()] + 1
                item[1]["curMovID"][pShip.GetName()] = iThisMovID
        
                fDuration = 200.0
                if item[1].has_key("WarpDuration"):
                        fDuration = item[1]["WarpDuration"]
                    
                # Rotation
                lStartingRotation = item[1]["currentRotation"][pShip.GetName()]
                lStoppingRotation = lStartingRotation
                if item[1].has_key("AttackRotation") and pShip.GetAlertLevel() == 2:
                                lStoppingRotation = item[1]["AttackRotation"]
		else:
			lStoppingRotation = item[1]["Rotation"]
                
                # Translation
                lStartingTranslation = item[1]["currentPosition"][pShip.GetName()]
                lStoppingTranslation = lStartingTranslation
                if item[1].has_key("AttackPosition") and pShip.GetAlertLevel() == 2:
                                lStoppingTranslation = item[1]["AttackPosition"]
		else:
			lStoppingTranslation = item[1]["Position"]
        
                iTime = 0.0
                iTimeNeededTotal = 0.0
                oMovingEvent = MovingEvent(pShip, item, fDuration, lStartingRotation, lStoppingRotation, lStartingTranslation, lStoppingTranslation)
                pSeq = App.TGSequence_Create()
                
		# do the move
                while(iTime < fDuration):
                        if iTime == 0.0:
                                iWait = 0.5 # we wait for the first run
                        else:
                                iWait = 0.01 # normal step
                        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "MovingAction", oMovingEvent), iWait)
                        iTimeNeededTotal = iTimeNeededTotal + iWait
                        iTime = iTime + 1
                pSeq.AppendAction(App.TGScriptAction_Create(__name__, "UpdateState", pShip, item, lStoppingRotation, lStoppingTranslation))
                pSeq.Play()
                
                # iLongestTime is for the part that needs the longest time...
                if iTimeNeededTotal > iLongestTime:
                        iLongestTime = iTimeNeededTotal
                
        # finally detach
        pSeq = App.TGSequence_Create()
        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "UpdateHardpointPositions", pShip, dHardpoints), iLongestTime)
        pSeq.AppendAction(App.TGScriptAction_Create(__name__, "WarpExitMoveFinishAction", pShip, pInstance, GetCurrentMoveID(pShip, pInstance)), 1.0)
        pSeq.Play()
        
        return 0
